import warnings
with warnings.catch_warnings():
	 warnings.simplefilter("ignore")
	 import networkit as nk
import networkx as nx

class DGraph(object):
	
	def __init__(self):
		
		self._GOR = None			# Object relation graph (edge = relation)
		self._GCR = None			# Class relation graph (edge = relation)
		self._GCM = None			# Class membership graph (edge = contains)
		
		self._GOR_names = {}		# {node_id: obj_id, ...}
		self._GOR_lookup = {}		# {obj_id: node_id, ...}
		self._GOR_data = {}  		# {obj_id: data, ...}
		self._GOR_labels = {}		# {(src_id, tgt_id): {label: weight, ...}, ...}
		
		self._GCR_names = {}		# {node_id: name, ...}
		self._GCR_lookup = {}		# {name: node_id, ...}
		self._GCR_data = {}  		# {name: data, ...}
		self._GCR_labels = {}		# {(src_id, tgt_id): [label, ...], ...}
		
		self._GCM_names = {}		# {node_id: name/obj_id, ...}
		self._GCM_lookup = {}		# {name/obj_id: node_id, ...}
		
		self.clear()
	
	def n_objects(self):
		
		return self._GOR.numberOfNodes()
	
	def n_classes(self):
		
		return self._GCR.numberOfNodes()
	
	def clear(self):
		
		self._GOR = nk.graph.Graph(directed = True)
		self._GCR = nk.graph.Graph(directed = True)
		self._GCM = nk.graph.Graph(directed = True)
		
		self._GOR_names.clear()
		self._GOR_lookup.clear()
		self._GOR_data.clear()
		self._GOR_labels.clear()
		
		self._GCR_names.clear()
		self._GCR_lookup.clear()
		self._GCR_data.clear()
		self._GCR_labels.clear()
		
		self._GCM_names.clear()
		self._GCM_lookup.clear()
	
	
	# ---- Objects
	# ------------------------------------------------------------------------
	
	def iter_objects(self):
		
		for obj_id in self._GOR_lookup:
			yield obj_id
	
	def iter_objects_data(self):
		
		for obj_id in self._GOR_data:
			yield self._GOR_data[obj_id]
	
	def iter_object_parents(self, obj_id):
		
		node_id = self._GCM_lookup[obj_id]
		for src in self._GCM.iterInNeighbors(node_id):
			yield self._GCM_names[src]
	
	def iter_object_ancestors(self, obj_id):
		
		node_id = self._GCM_lookup[obj_id]
		ancestors = set()
		queue = set([node_id])
		while queue:
			for src in self._GCM.iterInNeighbors(queue.pop()):
				if tgt not in ancestors:
					queue.add(tgt)
				ancestors.add(tgt)
		for node in ancestors:
			yield self._GCM_names[node]
	
	def has_object(self, obj_id):
		
		return obj_id in self._GOR_lookup
	
	def add_object(self, obj_id, data):
		
		if obj_id not in self._GOR_lookup:
			node_id = self._GOR.addNode()
			self._GOR_lookup[obj_id] = node_id
			self._GOR_names[node_id] = obj_id
			node_id = self._GCM.addNode()
			self._GCM_lookup[obj_id] = node_id
			self._GCM_names[node_id] = obj_id
		self._GOR_data[obj_id] = data
	
	def get_object_data(self, obj_id):
		
		return self._GOR_data[obj_id]
	
	def iter_object_relations(self, obj_id = None):
		
		if obj_id is None:
			for src, tgt in self._GOR.iterEdges():
				for label in self._GOR_labels[(src, tgt)]:
					yield self._GOR_names[src], self._GOR_names[tgt], label
		else:
			node_id = self._GOR_lookup[obj_id]
			for tgt in self._GOR.iterNeighbors(node_id):
				for label in self._GOR_labels[(node_id, tgt)]:
					yield self._GOR_names[tgt], label
	
	def add_object_relation(self, src_id, tgt_id, label, weight = None):
		
		src_id = self._GOR_lookup[src_id]
		tgt_id = self._GOR_lookup[tgt_id]
		
		if not self._GOR.hasEdge(src_id, tgt_id):
			self._GOR.addEdge(src_id, tgt_id)
			self._GOR_labels[(src_id, tgt_id)] = {}
		self._GOR_labels[(src_id, tgt_id)][label] = weight
	
	def has_object_relation(self, src_id, tgt_id, label):
		
		src_id = self._GOR_lookup[src_id]
		tgt_id = self._GOR_lookup[tgt_id]
		
		if not self._GOR.hasEdge(src_id, tgt_id):
			return False
		return label in self._GOR_labels[(src_id, tgt_id)]
	
	def get_object_relation_weight(self, src_id, tgt_id, label):
		
		src_id = self._GOR_lookup[src_id]
		tgt_id = self._GOR_lookup[tgt_id]
		
		if not self._GOR.hasEdge(src_id, tgt_id):
			return None
		return self._GOR_labels[(src_id, tgt_id)].get(label, None)
	
	def del_object_relation(self, src_id, tgt_id, label):
		
		src_id = self._GOR_lookup[src_id]
		tgt_id = self._GOR_lookup[tgt_id]
		
		if not self._GOR.hasEdge(src_id, tgt_id):
			return
		if label in self._GOR_labels[(src_id, tgt_id)]:
			del self._GOR_labels[(src_id, tgt_id)][label]
		if not self._GOR_labels[(src_id, tgt_id)]:
			self._GOR.removeEdge(src_id, tgt_id)
			del self._GOR_labels[(src_id, tgt_id)]
	
	def del_object(self, obj_id):
		
		node_id = self._GOR_lookup[obj_id]
		del self._GOR_names[node_id]
		del self._GOR_lookup[obj_id]
		del self._GOR_data[obj_id]
		for tgt in self._GOR.iterNeighbors(node_id):
			del self._GOR_labels[(node_id, tgt)]
		for src in self._GOR.iterInNeighbors(node_id):
			del self._GOR_labels[(src, node_id)]
		self._GOR.removeNode(node_id)
		
		node_id = self._GCM_lookup[obj_id]
		self._GCM.removeNode(node_id)
		del self._GCM_names[node_id]
		del self._GCM_lookup[obj_id]
	
	
	# ---- Classes
	# ------------------------------------------------------------------------
	
	def iter_classes(self):
		
		for name in self._GCR_lookup:
			yield name
	
	def iter_classes_data(self):
		
		for name in self._GCR_data:
			yield self._GCR_data[name]
	
	def has_class(self, name):
		
		return name in self._GCR_lookup
	
	def add_class(self, name, data):
		
		if name not in self._GCR_lookup:
			node_id = self._GCR.addNode()
			self._GCR_lookup[name] = node_id
			self._GCR_names[node_id] = name
			node_id = self._GCM.addNode()
			self._GCM_lookup[name] = node_id
			self._GCM_names[node_id] = name
		self._GCR_data[name] = data
	
	def rename_class(self, name, new_name):
		
		node_id = self._GCR_lookup[name]
		data = self._GCR_data[name]
		del self._GCR_lookup[name]
		del self._GCR_data[name]
		self._GCR_names[node_id] = new_name
		self._GCR_lookup[new_name] = node_id
		self._GCR_data[new_name] = data
		
		node_id = self._GCM_lookup[name]
		del self._GCM_lookup[name]
		self._GCM_names[node_id] = new_name
		self._GCM_lookup[new_name] = node_id
	
	def get_class_data(self, name):
		
		return self._GCR_data[name]
	
	def iter_class_children(self, name):
		
		node_id = self._GCM_lookup[name]
		for tgt in self._GCM.iterNeighbors(node_id):
			yield self._GCM_names[tgt]
	
	def iter_class_descendants(self, name):
		
		node_id = self._GCM_lookup[name]
		descendants = set()
		nk.traversal.Traversal.DFSfrom(self._GCM, node_id, lambda node: descendants.add(node))
		descendants.remove(node_id)
		for node in descendants:
			yield self._GCM_names[node]
	
	def iter_class_parents(self, child):
		
		node_id = self._GCM_lookup[child]
		for src in self._GCM.iterInNeighbors(node_id):
			yield self._GCM_names[src]
	
	def iter_class_ancestors(self, child):
		
		node_id = self._GCM_lookup[child]
		ancestors = set()
		queue = set([node_id])
		while queue:
			for src in self._GCM.iterInNeighbors(queue.pop()):
				if tgt not in ancestors:
					queue.add(tgt)
				ancestors.add(tgt)
		for node in ancestors:
			yield self._GCM_names[node]
	
	def iter_class_relations(self, name = None):
		
		if name is None:
			for src, tgt in self._GCR.iterEdges():
				for label in self._GCR_labels[(src, tgt)]:
					yield self._GCR_names[src], self._GCR_names[tgt], label
		else:
			node_id = self._GCR_lookup[name]
			for tgt in self._GCR.iterNeighbors(node_id):
				for label in self._GCR_labels[(node_id, tgt)]:
					yield self._GCR_names[tgt], label
	
	def add_class_child(self, name, child):
		# child = obj_id or subclass_name
		
		src_id = self._GCM_lookup[name]
		tgt_id = self._GCM_lookup[child]
		if not self._GCM.hasEdge(src_id, tgt_id):
			self._GCM.addEdge(src_id, tgt_id)
	
	def has_class_child(self, name, child):
		# child = obj_id or subclass_name
		
		src_id = self._GCM_lookup[name]
		tgt_id = self._GCM_lookup[child]
		
		return self._GCM.hasEdge(src_id, tgt_id)
	
	def has_class_relation(self, src_name, tgt_name, label):
		
		src_id = self._GCR_lookup[src_name]
		tgt_id = self._GCR_lookup[tgt_name]
		
		if not self._GCR.hasEdge(src_id, tgt_id):
			return False
		
		return label in self._GCR_labels[(src_id, tgt_id)]
	
	def add_class_relation(self, src_name, tgt_name, label):
		
		src_id = self._GCR_lookup[src_name]
		tgt_id = self._GCR_lookup[tgt_name]
		
		if not self._GCR.hasEdge(src_id, tgt_id):
			self._GCR.addEdge(src_id, tgt_id)
			self._GCR_labels[(src_id, tgt_id)] = []
		if label not in self._GCR_labels[(src_id, tgt_id)]:
			self._GCR_labels[(src_id, tgt_id)].append(label)
	
	def del_class_relation(self, src_name, tgt_name, label):
		
		src_id = self._GCR_lookup[src_name]
		tgt_id = self._GCR_lookup[tgt_name]
		
		if not self._GCR.hasEdge(src_id, tgt_id):
			return
		if label in self._GCR_labels[(src_id, tgt_id)]:
			self._GCR_labels[(src_id, tgt_id)].remove(label)
		if not self._GCR_labels[(src_id, tgt_id)]:
			self._GCR.removeEdge(src_id, tgt_id)
			del self._GCR_labels[(src_id, tgt_id)]
	
	def del_class_child(self, name, child):
		# child = obj_id or subclass_name
		
		src_id = self._GCM_lookup[name]
		tgt_id = self._GCM_lookup[child]
		if self._GCM.hasEdge(src_id, tgt_id):
			self._GCM.removeEdge(src_id, tgt_id)
	
	def del_class(self, name):
		
		node_id = self._GCR_lookup[name]
		del self._GCR_names[node_id]
		del self._GCR_lookup[name]
		del self._GCR_data[name]
		for tgt in self._GCR.iterNeighbors(node_id):
			del self._GCR_labels[(node_id, tgt)]
		for src in self._GCR.iterInNeighbors(node_id):
			del self._GCR_labels[(src, node_id)]
		self._GCR.removeNode(node_id)
		
		node_id = self._GCM_lookup[name]
		self._GCM.removeNode(node_id)
		del self._GCM_names[node_id]
		del self._GCM_lookup[name]
	
	
	# ---- Serialize
	# ------------------------------------------------------------------------
	
	def objects_to_nx(self):
		
		G = nx.MultiDiGraph()
		for obj_id in self._GOR_data:
			G.add_node(obj_id, data = self._GOR_data[obj_id])
		for src_id, tgt_id in self._GOR_labels:
			for label in self._GOR_labels[(src_id, tgt_id)]:
				weight = self._GOR_labels[(src_id, tgt_id)][label]
				if weight is None:
					G.add_edge(self._GOR_names[src_id], self._GOR_names[tgt_id], label)
				else:
					G.add_edge(self._GOR_names[src_id], self._GOR_names[tgt_id], label, weight = weight)
		return G
	
	def classes_to_nx(self):
		
		G = nx.MultiDiGraph()
		for name in self._GCR_data:
			G.add_node(name, data = self._GCR_data[name])
		for src_id, tgt_id in self._GCR_labels:
			for label in self._GCR_labels[(src_id, tgt_id)]:
				G.add_edge(self._GCR_names[src_id], self._GCR_names[tgt_id], label)
		return G
	
	def members_to_nx(self):
		
		G = nx.DiGraph()
		for name in self._GCM_lookup:
			G.add_node(name)
		for src_id, tgt_id in self._GCM.iterEdges():
			G.add_edge(self._GCM_names[src_id], self._GCM_names[tgt_id])
		return G
	
	def objects_from_nx(self, G):
		
		self._GOR = nk.graph.Graph(directed = True)
		self._GOR_names = {}		# {node_id: obj_id, ...}
		self._GOR_lookup = {}		# {obj_id: node_id, ...}
		self._GOR_data = {}  		# {obj_id: data, ...}
		self._GOR_labels = {}		# {(src_id, tgt_id): [[label, weight], ...], ...}
		for obj_id, data in G.nodes(data = "data"):
			node_id = self._GOR.addNode()
			self._GOR_lookup[obj_id] = node_id
			self._GOR_names[node_id] = obj_id
			self._GOR_data[obj_id] = data
		for src_id, tgt_id, label in G.edges(keys = True):
			weight = G.edges[src_id, tgt_id, label].get("weight", None)
			src_id = self._GOR_lookup[src_id]
			tgt_id = self._GOR_lookup[tgt_id]
			if not self._GOR.hasEdge(src_id, tgt_id):
				self._GOR.addEdge(src_id, tgt_id)
				self._GOR_labels[(src_id, tgt_id)] = {}
			self._GOR_labels[(src_id, tgt_id)][label] = weight
		
	def classes_from_nx(self, G):
		
		self._GCR = nk.graph.Graph(directed = True)
		self._GCR_names = {}		# {node_id: name, ...}
		self._GCR_lookup = {}		# {name: node_id, ...}
		self._GCR_data = {}  		# {name: data, ...}
		self._GCR_labels = {}		# {(src_id, tgt_id): [[label, weight], ...], ...}
		for name, data in G.nodes(data = "data"):
			node_id = self._GCR.addNode()
			self._GCR_lookup[name] = node_id
			self._GCR_names[node_id] = name
			self._GCR_data[name] = data
		for src_id, tgt_id, label in G.edges(keys = True):
			src_id = self._GCR_lookup[src_id]
			tgt_id = self._GCR_lookup[tgt_id]
			if not self._GCR.hasEdge(src_id, tgt_id):
				self._GCR.addEdge(src_id, tgt_id)
				self._GCR_labels[(src_id, tgt_id)] = []
			self._GCR_labels[(src_id, tgt_id)].append(label)
	
	def members_from_nx(self, G):
		
		self._GCM = nk.graph.Graph(directed = True)
		self._GCM_names = {}		# {node_id: name/obj_id, ...}
		self._GCM_lookup = {}		# {name/obj_id: node_id, ...}
		for name in G.nodes():
			node_id = self._GCM.addNode()
			self._GCM_lookup[name] = node_id
			self._GCM_names[node_id] = name
		for src_id, tgt_id in G.edges():
			src_id = self._GCM_lookup[src_id]
			tgt_id = self._GCM_lookup[tgt_id]
			self._GCM.addEdge(src_id, tgt_id)
		
	
	def objects_to_json(self, graph_attrs):
		
		return nx.node_link_data(self.objects_to_nx(), attrs = graph_attrs)
	
	def classes_to_json(self, graph_attrs):
		
		return nx.node_link_data(self.classes_to_nx(), attrs = graph_attrs)
	
	def members_to_json(self, graph_attrs):
		
		return nx.node_link_data(self.members_to_nx(), attrs = graph_attrs)
	
	def objects_from_json(self, data, graph_attrs):
		
		G = nx.node_link_graph(data, directed = True, multigraph = True, attrs = graph_attrs)
		self.objects_from_nx(G)
	
	def classes_from_json(self, data, graph_attrs):
		
		G = nx.node_link_graph(data, directed = True, multigraph = True, attrs = graph_attrs)
		self.classes_from_nx(G)
	
	def members_from_json(self, data, graph_attrs):
		
		G = nx.node_link_graph(data, directed = True, multigraph = False, attrs = graph_attrs)
		self.members_from_nx(G)
	
	
	def objects_to_pickle(self):
		
		return [
			self._GOR,
			self._GOR_names,
			self._GOR_lookup,
			self._GOR_data,
			self._GOR_labels,
		]
	
	def classes_to_pickle(self):
		
		return [
			self._GCR,
			self._GCR_names,
			self._GCR_lookup,
			self._GCR_data,
			self._GCR_labels,		
		]
	
	def members_to_pickle(self):
		
		return [
			self._GCM,
			self._GCM_names,
			self._GCM_lookup,
		]
	
	def objects_from_pickle(self, data):
		
		[
			self._GOR,
			self._GOR_names,
			self._GOR_lookup,
			self._GOR_data,
			self._GOR_labels,
		] = data
	
	def classes_from_pickle(self, data):
		
		[
			self._GCR,
			self._GCR_names,
			self._GCR_lookup,
			self._GCR_data,
			self._GCR_labels,		
		] = data
	
	def members_from_pickle(self, data):
		
		[
			self._GCM,
			self._GCM_names,
			self._GCM_lookup,
		] = data




from deposit.datasource.abstract_datasource import AbstractDatasource
from deposit.datasource.memory import Memory
from deposit.datasource.json import JSON
from deposit.datasource.pickle import Pickle
from deposit.datasource.db import DB
from deposit.datasource.db_rel import DBRel



from deposit.store.abstract_delement import AbstractDElement

class DClass(AbstractDElement):
	
	def __init__(self, store, name, order):
		
		AbstractDElement.__init__(self, store)
		
		self.name = name
		self.order = order
		
		self._descriptors = set()
		self._json_descriptors = None
		
	
	def _on_class_deleted(self, cls):
		
		if cls in self._descriptors:
			self._descriptors.remove(cls)
			return True
		return False
	
	def to_dict(self):
		
		data = dict(
			order = self.order
		)
		descriptors = [descr.name for descr in self._descriptors]
		if descriptors:
			data["descriptors"] = descriptors
		
		return data
	
	def from_dict_1(self, data):
		
		self._json_descriptors = data.get("descriptors", [])
		
		return self
	
	def from_dict_2(self):
		
		for name in self._json_descriptors:
			self._descriptors.add(self._store.get_class(name))
		self._json_descriptors = None
	
	
	def rename(self, new_name):
		
		if isinstance(new_name, DClass):
			new_name = new_name.name
		if self._store.has_class(new_name):
			raise Exception("Cannot rename class %s to %s, as it already exists." % (self.name, new_name))
		self._store.G.rename_class(self.name, new_name)
		self.name = new_name
		self._store.callback_changed([self])
	
	
	def set_descriptor(self, name):
		
		descr = self._store.add_class(name)
		if descr in self._descriptors:
			return
		self._descriptors.add(descr)
		self._store._on_descriptor_set(descr)
		self._store.callback_changed([self])
	
	def has_descriptors(self):
		
		return len(self._descriptors) > 0
	
	def has_descriptor(self, name):
		
		descr = self._store.get_class(name)
		if descr is None:
			return False
		return (descr in self._descriptors)
	
	def del_descriptor(self, name):
		
		descr = self._store.get_class(name)
		if descr is None:
			return
		if descr in self._descriptors:
			self._descriptors.remove(descr)
			self._store._on_descriptor_deleted()
			self._store.callback_changed([self])
	
	def get_descriptors(self, ordered = False):
		
		if ordered:
			return sorted(list(self._descriptors), key = lambda descr: descr.order)
		return list(self._descriptors)
	
	def get_descriptor_names(self, ordered = False):
		
		return [descr.name for descr in self.get_descriptors(ordered)]
	
	def get_object_descriptor_names(self, ordered = False, direct_only = False):
		
		descriptors = set()
		for obj in self.get_members(direct_only = direct_only):
			descriptors.update(obj.get_descriptors())
		if ordered:
			return [descr.name for descr in sorted(list(descriptors), key = lambda descr: descr.order)]
		return [descr.name for descr in descriptors]
	
	def add_member(self, obj_id = None):
		# obj_id = int or DObject or None (will be created)
		#
		# returns DObject
		
		if obj_id is None:
			obj_id = self._store.add_object().id
		elif obj_id.__class__.__name__ == "DObject":
			obj_id = obj_id.id
		self._store.check_obj_id(obj_id)
		self._store.G.add_class_child(self.name, obj_id)
		obj = self._store.G.get_object_data(obj_id)
		for descr in obj.get_descriptors():
			self.set_descriptor(descr)
		self._store.callback_changed([self, obj])
		
		return obj
	
	def get_members(self, direct_only = False):
		# direct_only = if True, don't return members of subclasses
		#
		# returns set([DObject, ...])
		
		objects = set()
		if direct_only:
			tgts = list(self._store.G.iter_class_children(self.name))
		else:
			tgts = list(self._store.G.iter_class_descendants(self.name))
		for tgt in tgts:
			if isinstance(tgt, int):
				objects.add(self._store.G.get_object_data(tgt))
		
		return objects
	
	def del_member(self, obj_id):
		
		if obj_id.__class__.__name__ == "DObject":
			obj_id = obj_id.id
		if self._store.G.has_class_child(self.name, obj_id):
			self._store.G.del_class_child(self.name, obj_id)
			obj = self._store.G.get_object_data(obj_id)
			self._store.callback_changed([self, obj])
	
	
	def add_subclass(self, name):
		
		if isinstance(name, DClass):
			name = name.name
		cls = self._store.add_class(name)
		self._store.G.add_class_child(self.name, name)
		self._store.callback_changed([self, cls])
		
		return cls
	
	def get_subclasses(self, ordered = False):
		# returns [DClass, ...]
		
		subclasses = set()
		for tgt in self._store.G.iter_class_descendants(self.name):
			if isinstance(tgt, str):
				subclasses.add(self._store.G.get_class_data(tgt))
		
		if ordered:
			return sorted(list(subclasses), key = lambda cls: cls.order)
		return list(subclasses)
	
	def get_superclasses(self, ordered = False, recursive = False):
		# returns [DClass, ...]
		
		superclasses = set()
		if recursive:
			for src in self._store.G.iter_class_ancestors(self.name):
				if isinstance(src, str):
					superclasses.add(self._store.G.get_class_data(src))
		else:
			for src in self._store.G.iter_class_parents(self.name):
				if isinstance(src, str):
					superclasses.add(self._store.G.get_class_data(src))
		
		if ordered:
			return sorted(list(superclasses), key = lambda cls: cls.order)
		return list(superclasses)
	
	def del_subclass(self, name):
		
		if isinstance(name, DClass):
			name = name.name
		if self._store.G.has_class_child(self.name, name):
			self._store.G.del_class_child(self.name, name)
			self._store.callback_changed([self, self._store.G.get_class_data(name)])
	
	
	def has_relation(self, name_tgt, label):
		
		if isinstance(name_tgt, DClass):
			name_tgt = name_tgt.name
		return self._store.G.has_class_relation(self.name, name_tgt, label)
	
	def add_relation(self, name_tgt, label):
		
		if isinstance(name_tgt, DClass):
			name_tgt = name_tgt.name
		name_src = self.name
		name_src, name_tgt, label = self._store._check_inverse_rel(name_src, name_tgt, label)
		self._store.G.add_class_relation(name_src, name_tgt, label)
		self._store.G.add_class_relation(name_tgt, name_src, "~" + label)
		self._store._on_relation_added(label)
		self._store.callback_changed([
			self._store.G.get_class_data(name_src),
			self._store.G.get_class_data(name_tgt),
		])
	
	def get_relations(self, name_tgt = None):
		# returns [(DClass, label), ...]
		
		if isinstance(name_tgt, DClass):
			name_tgt = name_tgt.name
		
		for tgt, label in self._store.G.iter_class_relations(self.name):
			if (name_tgt is not None) and (tgt != name_tgt):
				continue
			yield (self._store.G.get_class_data(tgt), label)
	
	def get_object_relations(self, class_tgt = None, direct_only = False):
		# get between objects of this class and other classes
		# returns {(DClass, label), ...}
		
		if isinstance(class_tgt, str):
			class_tgt = self._store.get_class(class_tgt)
		relations = set()
		for obj_src in self.get_members(direct_only = direct_only):
			for obj_tgt, label in obj_src.get_relations():
				classes_tgt = obj_tgt.get_classes()
				if class_tgt is not None:
					if class_tgt not in classes_tgt:
						continue
					relations.add((class_tgt, label))
				else:
					for cls in classes_tgt:
						relations.add((cls, label))
		
		return relations
	
	def del_relation(self, name_tgt, label):
		
		if isinstance(name_tgt, DClass):
			name_tgt = name_tgt.name
		name_src = self.name
		name_src, name_tgt, label = self._store._check_inverse_rel(name_src, name_tgt, label)
		
		if self._store.G.has_class_relation(name_src, name_tgt, label):
			self._store.G.del_class_relation(name_src, name_tgt, label)
			self._store.G.del_class_relation(name_tgt, name_src, "~" + label)
			self._store._on_relation_deleted()
			self._store.callback_changed([
				self._store.G.get_class_data(name_src),
				self._store.G.get_class_data(name_tgt),
			])
	
	def __repr__(self):
		
		return "DClass(%s)" % (self.name)




# -*- coding: utf-8 -*-
#!/usr/bin/env python
#

from setuptools import setup, find_packages
import pathlib

try:
	from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
	class bdist_wheel(_bdist_wheel):
		def finalize_options(self):
			_bdist_wheel.finalize_options(self)
			self.root_is_pure = False
except ImportError:
	bdist_wheel = None

here = pathlib.Path(__file__).parent.resolve()

long_description = (here / "README.md").read_text(encoding="utf-8")

setup(
	name="deposit",
	version="1.4.1",
	description="Graph database focused on scientific data collection.",
	long_description=long_description,
	long_description_content_type="text/markdown",
	url="https://github.com/demjanp/deposit",
	author="Peter Demján",
	author_email="peter.demjan@gmail.com",
	classifiers=[
		"Development Status :: 5 - Production/Stable",
		"Intended Audience :: Science/Research",
		"Topic :: Database",
		"License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
		"Programming Language :: Python :: 3.10",
		"Operating System :: Microsoft :: Windows :: Windows 10",
	],
	keywords="database, graph",
	package_dir={"": "src"},
	packages=find_packages(where="src"),
	python_requires=">=3.10, <4",
	install_requires=[
		'natsort>=8.1.0',
		'networkit>=10.0',
		'networkx>=2.6.3',
		'openpyxl>=3.0.10',
		'Pillow>=9.1.1',
		'psycopg2>=2.9.3',
		'pyshp>=2.3.0',
		'PySide2>=5.15.2.1',
		'Shapely>=1.8.2',
		'Unidecode>=1.3.4',
		'validators>=0.20.0',
	],
	cmdclass={'bdist_wheel': bdist_wheel},
)



from deposit.externalsource.abstract_externalsource import AbstractExternalsource

from openpyxl.styles import (Font)
from openpyxl import (load_workbook, Workbook)

class XLSX(AbstractExternalsource):
	
	def __init__(self):
		
		AbstractExternalsource.__init__(self)
		
		self._columns = {} # {sheet: {column_idx: name, ...}, ...}
		self._data = {} # {sheet: [row, ...], ...}; row = {column_idx: value, ...}
	
	def sheets(self):
		
		return list(self._columns.keys())
	
	def column_count(self, sheet):
		
		if sheet in self._columns:
			return len(self._columns[sheet])
		return 0
	
	def row_count(self, sheet):
		
		if sheet not in self._data:
			return 0
		return len(self._data[sheet])
	
	def column_name(self, sheet, column_idx):
		
		if sheet not in self._data:
			return None
		if column_idx not in self._columns[sheet]:
			return None
		return self._columns[sheet][column_idx]
	
	def data(self, sheet, row_idx, column_idx):
		
		if sheet not in self._data:
			return None
		if row_idx > len(self._data[sheet]) - 1:
			return None
		if column_idx not in self._data[sheet][row_idx]:
			return None
		return self._data[sheet][row_idx][column_idx]
	
	def save(self, get_header, get_item, n_rows, n_cols, path = None, url = None, *args, **kwargs):
		# get_header(col, user_role = False) = column_name or (class_name, descriptor_name) if user_role = True
		# get_item(row, col) = deposit_gui.QueryItem
		
		self.set_path(path, url)
		path = self.get_path()
		if path is None:
			return False
		
		wb = Workbook()
		wb.guess_types = True
		ws = wb.active
		ws.append(self.get_header_data(get_header, n_cols))
		for col in range(n_cols):
			ws.cell(row = 1, column = col + 1).font = Font(bold = True)
		for row in range(n_rows):
			ws.append(self.get_row_data(row, get_item, n_cols))
		wb.save(path)
		
		return True
	
	def load(self, path = None, url = None, *args, **kwargs):
		
		self.set_path(path, url)
		path = self.get_path()
		if path is None:
			return False
		
		self._columns.clear()
		self._data.clear()
		
		wb = load_workbook(filename = path, read_only = True)
		for sheet in wb.sheetnames:
			self._columns[sheet] = {}
			self._data[sheet] = []
			ws = wb[sheet]
			idx = 0
			idx_lookup = {} # {i: idx, ...}
			for i, cell in enumerate(list(ws.iter_rows(max_row = 1))[0]):
				value = cell.value
				if value is not None:
					value = str(value).strip()
				if value:
					self._columns[sheet][idx] = value
					idx_lookup[i] = idx
					idx += 1
			if not self._columns[sheet]:
				continue
			for row in ws.iter_rows(min_row = 2):
				self._data[sheet].append({})
				for i, cell in enumerate(row):
					value = cell.value
					if value is not None:
						value = str(value).strip()
					if value and (i in idx_lookup):
						self._data[sheet][-1][idx_lookup[i]] = value
		
		return True
		
	
		




from deposit.store.dgraph import DGraph
from deposit.store.dobject import DObject
from deposit.store.dclass import DClass
from deposit.store.dresource import DResource
from deposit import datasource as Datasource
from deposit.query.query import Query

from deposit.utils.fnc_files import (
	as_url, url_to_path, extract_filename, get_image_format, delete_stored, 
	store_locally, update_image_filename, get_updated_local_url, open_url, 
	clear_temp_dir, prune_local_files
)

from collections import defaultdict
from natsort import natsorted
import os

class Store(object):
	
	def __init__(self, keep_temp = False, *args, **kwargs):
		
		self.G = DGraph()
		
		self._datasource = Datasource.Memory()
		self._resources = {}			# {url: DResource, ...}
		self._user_tools = []
		self._queries = {}
		self._local_folder = None		# folder to store DResource files flagged as is_stored
		self._keep_temp = keep_temp
		self._saved = True
		
		self._min_free_id = None
		self._max_order = 0
		self._descriptors = None
		self._relation_labels = None
		
		self._callback_added = set()
		self._callback_deleted = set()
		self._callback_changed = set()
		self._callback_saved = set()
		self._callback_loaded = set()
		self._callback_local_folder_changed = set()
		self._callback_user_tools_changed = set()
		self._callback_queries_changed = set()
		
	def _init_min_free_id(self) -> None:
		
		if self._min_free_id is None:
			self._min_free_id = 1
			while self.G.has_object(self._min_free_id):
				self._min_free_id += 1
	
	def _check_inverse_rel(self, src, tgt, label: str) -> tuple:
		
		if label.startswith("~"):
			label = label.strip("~")
			return tgt, src, label
		
		return src, tgt, label
	
	def _on_descriptor_set(self, descr):
		
		if self._descriptors is not None:
			self._descriptors.add(descr)
	
	def _on_descriptor_deleted(self):
		
		self._descriptors = None
	
	def _on_relation_added(self, label):
		
		if self._relation_labels is not None:
			self._relation_labels.add(label)
	
	def _on_relation_deleted(self):
		
		self._relation_labels = None
	
	def _on_resource_deleted(self, resource):
		
		if not resource.object_ids:
			self.del_resource(resource)
	
	
	# ---- Callbacks
	# ------------------------------------------------------------------------
	def set_callback_added(self, func):
		# func(elements: list); elements = [DObject, DClass, ...]
		
		self._callback_added.add(func)
	
	def callback_added(self, elements):
		
		self._saved = False
		for func in self._callback_added:
			func(elements)
	
	def set_callback_deleted(self, func):
		# func(elements: list); elements = [obj_id, name, ...]
		
		self._callback_deleted.add(func)
	
	def callback_deleted(self, elements):
		
		self._saved = False
		for func in self._callback_deleted:
			func(elements)
	
	def set_callback_changed(self, func):
		# func(elements: list); elements = [DObject, DClass, ...]
		
		self._callback_changed.add(func)
	
	def callback_changed(self, elements):
		
		self._saved = False
		for func in self._callback_changed:
			func(elements)
	
	def set_callback_saved(self, func):
		# func(datasource)
		
		self._callback_saved.add(func)
	
	def callback_saved(self, datasource):
		
		self._saved = True
		for func in self._callback_saved:
			func(datasource)
	
	def set_callback_loaded(self, func):
		# func()
		
		self._callback_loaded.add(func)
	
	def callback_loaded(self):
		
		self._saved = True
		for func in self._callback_loaded:
			func()
	
	def set_callback_local_folder_changed(self, func):
		# func()
		
		self._callback_local_folder_changed.add(func)
	
	def callback_local_folder_changed(self):
		
		self._saved = False
		for func in self._callback_local_folder_changed:
			func()
	
	def set_callback_queries_changed(self, func):
		# func()
		
		self._callback_queries_changed.add(func)
	
	def callback_queries_changed(self):
		
		self._saved = False
		for func in self._callback_queries_changed:
			func()
	
	def set_callback_user_tools_changed(self, func):
		# func()
		
		self._callback_user_tools_changed.add(func)
	
	def callback_user_tools_changed(self):
		
		self._saved = False
		for func in self._callback_user_tools_changed:
			func()
	
	
	# ---- General
	# ------------------------------------------------------------------------
	def clear(self) -> None:
		
		self.G.clear()
		
		self._resources = {}
		self._user_tools = []
		self._queries = {}
		self._local_folder = None
		self._saved = False
		
		self._min_free_id = None
		self._max_order = 0
		self._descriptors = None
		self._relation_labels = None
	
	def get_updated_url(self, resource):
		
		if not resource.is_stored:
			return resource.url
		url_new = get_updated_local_url(resource.url, self.get_folder())
		if url_new != resource.url:
			resource.value = (url_new, resource.filename, resource.is_stored, resource.is_image)
		return url_new
	
	def set_local_folder(self, path, progress = None):
		
		if path is None:
			self._local_folder = None
			self.callback_local_folder_changed()
			return
		if not os.path.isdir(path):
			os.makedirs(path, exist_ok = True)
		self._local_folder = os.path.normpath(os.path.abspath(path))
		
		cmax = len(self.get_objects())
		cnt = 1
		to_del = set()
		for obj in self.get_objects():
			if progress is not None:
				progress.update_state(text = "Copying Resources", value = cnt, maximum = cmax)
			for descr in obj.get_descriptors():
				resource = obj.get_descriptor(descr)
				if isinstance(resource, DResource):
					url, filename, is_stored, is_image = resource.value
					if is_stored:
						url_new = get_updated_local_url(url, self.get_folder())
						if url_new is None:
							url_new, is_stored = self.store_resource(url, filename)
						resource.value = (url_new, filename, is_stored, is_image)
						self._resources[url_new] = resource
						to_del.add(url)
		for url in to_del:
			if url in self._resources:
				del self._resources[url]
		if progress is not None:
			progress.stop()
		
		self.callback_local_folder_changed()
	
	def has_local_folder(self):
		
		return (self._local_folder is not None)
	
	def get_folder(self):
		
		if self._local_folder is None:
			if self._datasource is None:
				return None
			return self._datasource.get_folder()
		return self._local_folder
	
	def get_resource_urls(self):
		
		return set(self._resources.keys())
	
	def store_resource(self, url: str, filename: str) -> tuple[str, bool]:
		# if possible, store resource locally
		#
		# returns (url, is_stored); url = new url if stored locally; is_stored = True/False
		
		folder = self.get_folder()
		if folder is None:
			return (url, False)
		
		url_new = store_locally(url, filename, folder, self.get_resource_urls())
		if url_new is None:
			return (url, False)
		return (url_new, True)
	
	def prune_resources(self):
		
		paths = set()
		for url in self._resources:
			url = get_updated_local_url(url, self.get_folder())
			if url is not None:
				paths.add(url_to_path(url))
		prune_local_files(paths, self.get_folder())
	
	def open_resource(self, resource):
		# resource = DResource
		
		url, filename, is_stored, is_image = resource.value
		if is_stored:
			url_new = get_updated_local_url(url, self.get_folder())
			if url_new is None:
				raise Exception("File %s not found" % (url))
			if url_new != url:
				resource.value = (url_new, filename, is_stored, is_image)
				url = url_new
		
		return open_url(url)
	
	
	# ---- Create
	# ------------------------------------------------------------------------
	def add_object(self) -> DObject:
		
		self._init_min_free_id()
		obj_id = self._min_free_id
		
		obj = DObject(self, obj_id)
		self.G.add_object(obj_id, obj)
		
		while self.G.has_object(self._min_free_id):
			self._min_free_id += 1
		
		self.callback_added([obj])
		
		return obj
	
	def add_class(self, name) -> DClass:
		
		if isinstance(name, DClass):
			name = name.name
		if not self.G.has_class(name):
			cls = DClass(self, name, self._max_order)
			self.G.add_class(name, cls)
			self._max_order += 1
			
			self.callback_added([cls])
		
		return self.G.get_class_data(name)
	
	def add_resource(self, url: str, filename: str = None, is_stored: bool = None, is_image: bool = None) -> DResource:
		# filename = "name.ext"; if None, determine automatically
		# is_stored = True if resource is stored in local folder
		#	= False, not stored locally
		#	= None, attempt to store it locally if possible & set is_stored flag
		# is_image = True if resource is an image; if None, determine automatically
		
		url = as_url(url)
		
		format = None
		if is_image is None:
			format = get_image_format(url)
			is_image = (format is not None)
		
		if filename is None:
			filename = extract_filename(url)
			if is_image:
				filename = update_image_filename(filename, format)
		
		if is_stored is None:
			is_stored = False
			if self.get_folder():
				url, is_stored = self.store_resource(url, filename)
		
		if url not in self._resources:
			self._resources[url] = DResource((url, filename, is_stored, is_image))
		
		return self._resources[url]
	
	def add_user_tool(self, user_tool):
		
		self._user_tools.append(user_tool)
		self.callback_user_tools_changed()
	
	def add_saved_query(self, title, querystr):
		
		if not (title and querystr):
			return
		self._queries[title] = querystr
		self.callback_queries_changed()
	
	def find_object_with_descriptors(self, classes, data, locations = {}):
		# classes = [DClass, None, ...]
		# data = {descriptor_name: value, ...}
		# locations = {descriptor_name: location, ...}
		#
		# if found object has missing location, it is updated if a location is
		# supplied via locations
		
		def _get_class_members(cls):
			
			if cls is None:
				return set([obj for obj in self.get_objects() if obj.get_classes()])
			return cls.get_members(direct_only = True)
		
		def _check_location(obj, descriptor_name, locations):
			
			if descriptor_name not in locations:
				return True
			location = obj.get_location(descriptor_name)
			if location is None:
				obj.set_location(descriptor_name, locations[descriptor_name])
				return True
			if location == locations[descriptor_name]:
				return True
			return False
		
		for obj in set.union(*[_get_class_members(cls) for cls in classes]):
			found = True
			for descriptor_name in data:
				if (obj.get_descriptor(descriptor_name) != data[descriptor_name]) and \
				(_check_location(obj, descriptor_name, locations)):
					found = False
					break
			if found:
				return obj
		
		return None
	
	def add_external_descriptor(self, obj, name, value):
		# localize descriptor if it is a DResource
		
		if isinstance(value, DResource):
			is_stored = value.is_stored
			if is_stored == True:
				is_stored = None
			obj.set_resource_descriptor(name, value.url, value.filename, is_stored, value.is_image)
		else:
			obj.set_descriptor(name, value)
	
	def add_object_with_descriptors(self, cls, data, locations = {}, obj = None):
		# cls = DClass or None
		# data = {descriptor_name: value, ...}
		# locations = {descriptor_name: location, ...}
		
		if obj is None:
			obj = self.add_object()
		if cls is not None:
			cls.add_member(obj.id)
		for descriptor_name in data:
			value = data[descriptor_name]
			self.add_external_descriptor(obj, descriptor_name, value)
			if descriptor_name in locations:
				obj.set_location(descriptor_name, locations[descriptor_name])
		
		return obj
	
	def add_data_row(self, 
		data: dict, 
		relations: set = set(), 
		unique: set = set(), 
		existing = {}, 
		return_added = False,
	):
		# add multiple objects with classes at once & automatically add relations 
		#	based on class relations or as specified in the relations attribute
		# data = {(Class name, Descriptor name): value, ...}
		# relations = {(Class name 1, label, Class name 2), ...}
		# unique = {Class name, ...}; always add a new object to classes 
		#	specified here, otherwise re-use objects with identical descriptors
		# existing = {Class name: Object, ...}
		#	use existing object for specified classes (i.e. just update descriptors)
		#
		# returns n_added or (n_added, added) if return_added == True
		#	added = {Class name: Object, ...}
		
		collect = defaultdict(dict)
		for key in data:
			if data[key] is None:
				continue
			class_name, descriptor_name = key
			if (class_name is None) or (descriptor_name is None):
				continue
			cls = self.add_class(class_name)
			descr = self.add_class(descriptor_name)
			collect[cls][descriptor_name] = data[key]
		data = collect  # {cls: {descriptor_name: value, ...}, ...}
		
		collect = defaultdict(set)
		for class_name1, label, class_name2 in relations:
			cls1 = self.get_class(class_name1)
			cls2 = self.get_class(class_name2)
			if (cls1 is None) or (cls2 is None) or (label is None):
				continue
			if cls1 in data:
				collect[cls1].add((label, cls2))
			if cls2 in data:
				if label.startswith("~"):
					label = label[1:]
				else:
					label = "~" + label
				collect[cls2].add((label, cls1))
		for cls1 in data:
			for cls2, label in cls1.get_relations():
				if cls2 in data:
					collect[cls1].add((label, cls2))
		relations_all = collect  # {cls1: [(label, cls2), ...], ...}
		
		n_added = 0
		added = {}  # {cls: obj, ...}
		for cls in data:
			added[cls] = None
			if cls.name in existing:
				added[cls] = self.add_object_with_descriptors(
					cls, data[cls], obj = existing[cls.name]
				)
			elif cls.name not in unique:
				added[cls] = self.find_object_with_descriptors([cls], data[cls])
			if added[cls] is None:
				added[cls] = self.add_object_with_descriptors(cls, data[cls])
				n_added += 1
		
		for cls1 in relations_all:
			if cls1 not in added:
				continue
			for label, cls2 in relations_all[cls1]:
				if cls2 not in added:
					continue
				added[cls1].add_relation(added[cls2], label)
		
		if return_added:
			return n_added, dict([(cls.name, added[cls]) for cls in added])
		return n_added
	
	def import_store(self, store, unique: set = set(), progress = None) -> None:
		# unique = {Class name, ...}; always add a new object to classes 
		#	specified here, otherwise re-use objects with identical descriptors
		# progress = DProgress
		
		for cls in store.get_classes(ordered = True):
			self.add_class(cls.name)
		for cls in store.get_classes():
			cls_ = self.get_class(cls.name)
			for descr in cls.get_descriptors(ordered = True):
				cls_.set_descriptor(descr.name)
			for subcls in cls.get_subclasses():
				cls_.add_subclass(subcls)
			for cls2, label in cls.get_relations():
				cls_.add_relation(cls2.name, label)
		
		objects = store.get_objects()
		cmax = 2*len(objects)
		cnt = 1
		added = {}  # {orig_obj: new_obj, ...}
		for obj in objects:
			if progress is not None:
				if progress.cancel_pressed():
					return
				progress.update_state(text = "Importing", value = cnt, maximum = cmax)
				cnt += 1
			data = {}
			locations = {}
			for descr in obj.get_descriptors():
				data[descr.name] = obj.get_descriptor(descr)
				location = obj.get_location(descr)
				if location is not None:
					locations[descr.name] = location
			classes = [self.get_class(cls.name) for cls in obj.get_classes()]
			if not classes:
				classes = [None]
			obj_ = None
			if True not in [(cls.name in unique) for cls in classes if cls is not None]:
				obj_ = self.find_object_with_descriptors(classes, data, locations)
			if obj_ is None:
				obj_ = self.add_object_with_descriptors(classes[0], data, locations)
			for cls in classes:
				if cls is None:
					continue
				cls.add_member(obj_.id)
			added[obj] = obj_
		
		for obj in added:
			if progress is not None:
				if progress.cancel_pressed():
					return
				progress.update_state(text = "Importing", value = cnt, maximum = cmax)
				cnt += 1
			obj_ = added[obj]
			for obj2, label in obj.get_relations():
				weight = obj.get_relation_weight(obj2, label)
				obj_.add_relation(added[obj2], label, weight = weight)
	
	
	# ---- Read
	# ------------------------------------------------------------------------
	def get_object(self, obj_id) -> DObject:
		
		if isinstance(obj_id, DObject):
			obj_id = obj_id.id
		if self.G.has_object(obj_id):
			return self.G.get_object_data(obj_id)
		return None
	
	def get_objects(self) -> set:
		# return {DObject, ...}
		
		return set(obj for obj in self.G.iter_objects_data())
	
	def get_object_ids(self) -> set:
		
		return set(self.G.iter_objects())
	
	def check_obj_id(self, obj_id) -> None:
		
		if isinstance(obj_id, DObject):
			obj_id = obj_id.id
		if not self.G.has_object(obj_id):
			raise Exception("Specified object %s does not exist." % (str(obj_id)))
	
	
	def get_class(self, name) -> DClass:
		
		if isinstance(name, DClass):
			name = name.name
		if self.G.has_class(name):
			return self.G.get_class_data(name)
		return None
	
	def get_classes(self, ordered = False) -> list:
		# return [DClass, ...]
		
		classes = []
		for cls in self.G.iter_classes_data():
			classes.append(cls)
		if ordered:
			return sorted(classes, key = lambda cls: cls.order)
		return classes
	
	def get_class_names(self, ordered = False) -> list:
		# return [name, ...]
		
		if ordered:
			return [cls.name for cls in self.get_classes(ordered = True)]
		return list(self.G.iter_classes())
	
	def has_class(self, name) -> bool:
		
		if isinstance(name, DClass):
			name = name.name
		return self.G.has_class(name)
	
	
	def get_descriptors(self, ordered = False) -> list:
		# return [DClass, ...]
		
		if self._descriptors is None:
			self._descriptors = set()
			for obj in self.get_objects():
				for name in obj.get_descriptor_names():
					self._descriptors.add(self.G.get_class_data(name))
			for cls in self.get_classes():
				for name in cls.get_descriptor_names():
					self._descriptors.add(self.G.get_class_data(name))
		
		if ordered:
			return sorted(list(self._descriptors), key = lambda cls: cls.order)
		return list(self._descriptors)
	
	def get_descriptor_names(self, ordered = False) -> list:
		# return [name, ...]
		
		return [cls.name for cls in self.get_descriptors(ordered)]
	
	
	def get_user_tools(self):
		
		return self._user_tools.copy()
	
	
	def get_saved_query(self, title):
		
		if title in self._queries:
			return self._queries[title]
		return ""
	
	def get_saved_queries(self):
		
		return natsorted(self._queries.keys())
	
	
	def get_relation_labels(self) -> set:
		
		if self._relation_labels is None:
			self._relation_labels = set()
			for _, _, label in self.G.iter_object_relations():
				if not label.startswith("~"):
					self._relation_labels.add(label)
			for _, _, label in self.G.iter_class_relations():
				if not label.startswith("~"):
					self._relation_labels.add(label)
		
		return self._relation_labels
	
	
	def get_class_connections(self, classes: list, relations: list, progress = None) -> list:
		# classes = [class_name, ...]
		# relations = [(class1, label, class2), ...]
		#
		# returns [path, ...]; where path = [obj_id, ...]
		# 	first obj in path is a member of class_name from the first select
		# at least one class must be specified
		# if relations are specified, only accept connections between 
		# 	[class1] and [class2] if they are labeled [label]
		
		def get_class_members(class_name):
			
			if class_name == "*":
				return set([obj.id for obj in self.get_objects()])
			if class_name == "!*":
				return set([obj.id for obj in self.get_objects() if not obj.has_class()])
			if isinstance(class_name, tuple):
				return set(list(class_name))
			cls = self.get_class(class_name)
			if cls is None:
				return set()
			return set([obj.id for obj in cls.get_members()])
		
		def get_paths(src, objects2, relation_objs, n_relations, class_objs, n_classes, has_asterisk_label):
			
			if (not objects2) and (not relation_objs):
				return set([(src,)])
			
			queue = [[[src], set(), class_objs.get(src, set()).copy()]]  
			# queue = [[path, found_relations, found_classes], ...]
			done = set()
			while queue:
				path, found_rels, found_clss = queue.pop()
				found = False
				src_clss = class_objs.get(path[-1], set())
				for tgt, label in self.G.iter_object_relations(path[-1]):
					if tgt in path:
						continue
					idx = relation_objs.get((path[-1], label, tgt), None) if relation_objs else 0
					if (idx is None) and has_asterisk_label:
						idx = relation_objs.get((path[-1], "*", tgt), None)
					if idx is None:
						continue
					tgt_clss = class_objs.get(tgt, set())
					found_clss.update(tgt_clss)
					is_same_class = (len(tgt_clss.intersection(src_clss)) > 0)
					if len(found_clss) >= n_classes:
						done.add(tuple(sorted(path + ([] if is_same_class else [tgt]))))
					else:
						found_rels.add(idx)
						queue.append([path + [tgt], found_rels, found_clss])
					found = True
				if (not found) and (len(found_rels) >= n_relations):
					done.add(tuple(sorted(path)))
			
			return done
		
		relation_objs = {}
		has_asterisk_label = False
		for idx in range(len(relations)):
			cls1, label, cls2 = relations[idx]
			if (cls1, label, cls2) == ("*","*","*"):
				continue
			objects1 = get_class_members(cls1)
			objects2 = get_class_members(cls2)
			if label.startswith("~"):
				rev_label = label[1:]
			elif label == "*":
				has_asterisk_label = True
				rev_label = "*"
			else:
				rev_label = "~" + label
			for src in objects1:
				for tgt in objects2:
					relation_objs[(src, label, tgt)] = idx
					relation_objs[(tgt, rev_label, src)] = idx
		n_relations = len(relations)
		
		classes = [class_name for class_name in classes if \
			(self.has_class(class_name) or \
			isinstance(class_name, tuple) or \
			(class_name in ["*", "!*"]))]
		class_objs = defaultdict(set)
		for class_name in classes:
			for obj_id in get_class_members(class_name):
				class_objs[obj_id].add(class_name)
		n_classes = len(classes)
		
		paths = set()
		objects1 = get_class_members(classes[0])
		objects2 = set()
		for class_name in classes[1:]:
			objects2.update(get_class_members(class_name))
		cmax = len(objects1)
		cnt = 1
		for obj_id in objects1:
			if (progress is not None) and (cnt % 1000 == 0):
				if progress.cancel_pressed():
					return []
				progress.update_state(value = cnt, maximum = cmax)
			cnt += 1
			found = False
			paths_ = get_paths(obj_id, objects2, relation_objs, n_relations, class_objs, n_classes, has_asterisk_label)
			if paths_:
				paths.update(paths_)
			elif not relations:
				paths.add((obj_id,))
		
		if progress is not None:
			progress.update_state(value = cmax, maximum = cmax)
		
		return paths
	
	
	def get_query(self, querystr: str, **kwargs) -> object:
		
		return Query(self, querystr, **kwargs)
	
	
	def get_subgraph(self, objects: list, progress = None):
		# returns store which contains specified objects with their descriptors
		# and objects related to them, except objects from the same class as the
		# specified objects
		
		def get_or_add_object(store, obj, obj_lookup):
			
			if obj.id in obj_lookup:
				return obj_lookup[obj.id]
			obj2 = store.add_object()
			obj_lookup[obj.id] = obj2
			return obj2
		
		store = Store(keep_temp = True)
		
		# add objects
		obj_lookup = {}  # {orig_id: DObject in store, ...}
		classes0 = set()
		for obj in objects:
			obj_id = get_or_add_object(store, obj, obj_lookup)
			for cls in obj.get_classes():
				classes0.add(cls.name)
		
		# add object relations
		queue = set(obj_lookup.keys())
		done = set()
		classes_all = classes0.copy()
		while queue:
			orig_id = queue.pop()
			orig = self.get_object(orig_id)
			for orig2, label in orig.get_relations():
				classes = set([cls.name for cls in orig2.get_classes()])
				if classes.intersection(classes0):
					continue
				classes_all.update(classes)
				weight = orig.get_relation_weight(orig2, label)
				obj2 = get_or_add_object(store, orig2, obj_lookup)
				obj_lookup[orig_id].add_relation(obj2, label, weight = weight)
				if orig2.id in done:
					continue
				queue.add(orig2.id)
				done.add(orig2.id)
		
		# collect descriptor names
		for orig_id in obj_lookup:
			for descr in self.get_object(orig_id).get_descriptors():
				classes_all.add(descr.name)
		for class_name in classes_all.copy():
			for descr in self.get_class(class_name).get_descriptors():
				classes_all.add(descr.name)
		
		# add classes & members
		for cls in self.get_classes(ordered = True):
			if cls.name not in classes_all:
				continue
			cls_ = store.add_class(cls.name)
			for obj in cls.get_members():
				if obj.id not in obj_lookup:
					continue
				cls_.add_member(obj_lookup[obj.id].id)
		
		# add subclasses, descriptors & relations
		for class_name in classes_all:
			cls_orig = self.get_class(class_name)
			cls = store.get_class(class_name)
			for cls2 in cls_orig.get_subclasses():
				if cls2.name in classes_all:
					cls.add_subclass(cls2.name)
			for descr in cls_orig.get_descriptors():
				if descr.name in classes_all:
					cls.set_descriptor(descr.name)
			for cls2_orig, label in cls_orig.get_relations():
				if cls2_orig.name in classes_all:
					cls.add_relation(cls2_orig.name, label)
		
		# add descriptors & locations
		for orig_id in obj_lookup:
			orig = self.get_object(orig_id)
			obj = obj_lookup[orig_id]
			for descr in orig.get_descriptors():
				value = orig.get_descriptor(descr)
				self.add_external_descriptor(obj, descr.name, value)
				location = orig.get_location(descr)
				if location is not None:
					obj.set_location(descr.name, location)
		
		return store
	
	
	# ---- Update
	# ------------------------------------------------------------------------
	def switch_order(self, cls1, cls2) -> None:
		
		if isinstance(cls1, str):
			cls1 = self.get_class(cls1)
		if isinstance(cls2, str):
			cls2 = self.get_class(cls2)
		if (cls1 is None) or (cls2 is None):
			raise Exception("Invalid classes specified")
		
		cls1.order, cls2.order = cls2.order, cls1.order
		
		self.callback_changed([cls1, cls2])
	
	
	# ---- Delete
	# ------------------------------------------------------------------------
	def del_object(self, obj_id) -> None:
		
		if isinstance(obj_id, DObject):
			obj_id = obj_id.id
		if not self.G.has_object(obj_id):
			return
		if self.get_object(obj_id).has_descriptors():
			self._descriptors = None
		self.G.del_object(obj_id)
		self._init_min_free_id()
		if obj_id < self._min_free_id:
			self._min_free_id = obj_id
		self.callback_deleted([obj_id])
	
	def del_class(self, name) -> None:
		
		if isinstance(name, DClass):
			name = name.name
		if not self.G.has_class(name):
			return
		# check and remove if class is descriptor for objects / classes
		descr = self.G.get_class_data(name)
		deleted = False
		for obj in self.G.iter_objects_data():
			if obj._on_class_deleted(descr):
				deleted = True
		for cls in self.G.iter_classes_data():
			if cls._on_class_deleted(descr):
				deleted = True
		if deleted:
			self._descriptors = None
		self.G.del_class(name)
		self._max_order = 0
		classes = self.get_classes()
		if classes:
			self._max_order = max([cls.order for cls in classes]) + 1
		self.callback_deleted([name])
	
	def del_resource(self, resource: DResource) -> None:
		
		if resource.is_stored:
			delete_stored(resource.url, self.get_folder())
		url = resource.url
		if url in self._resources:
			del self._resources[url]
	
	def del_user_tool(self, label):
		
		self._user_tools = [user_tool for user_tool in self._user_tools if user_tool["label"] != label]
		self.callback_user_tools_changed()
	
	def del_saved_query(self, title):
		
		if title in self._queries:
			del self._queries[title]
			self.callback_queries_changed()
	
	
	# ---- Persistence
	# ------------------------------------------------------------------------
	def init_datasource(self, format, kwargs):
		# format = Datasource or format
		
		ext_to_format = {
			".json": "JSON",
			".pickle": "Pickle",
		}
		
		if format is None:
			ext = None
			for key in ["path", "url"]:
				if kwargs.get(key, None) is not None:
					ext = os.path.splitext(kwargs[key])[-1]
					break
			if ext is not None:
				ext = ext.lower()
				if ext in ext_to_format:
					format = ext_to_format[ext]
		if format is None:
			if ("identifier" in kwargs) and ("connstr" in kwargs):
				for datasource in [Datasource.DB, Datasource.DBRel]:
					ds = datasource()
					ds.set_connstr(kwargs["connstr"])
					ds.set_identifier(kwargs["identifier"])
					if ds.is_valid():
						return ds
					ds = None
				ds = None
			
		if format is None:
			format = self._datasource
		
		if format is None:
			raise Exception("Format not specified")
		
		if isinstance(format, Datasource.AbstractDatasource):
			return format
		if not hasattr(Datasource, format):
			raise Exception("Invalid format specified: %s" % (format))
		return getattr(Datasource, format)()
	
	def get_datasource(self):
		
		return self._datasource
	
	def set_datasource(self, datasource):
		
		self._datasource = datasource
	
	def save(self, datasource = None, *args, **kwargs):
		
		# datasource = Datasource or format
		datasource = self.init_datasource(datasource, kwargs)
		if datasource.save(self, datasource = datasource, *args, **kwargs):
			self.callback_saved(datasource)
			return True
		return False
	
	def load(self, datasource = None, *args, **kwargs):
		# datasource = Datasource or format
		
		datasource = self.init_datasource(datasource, kwargs)
		if datasource.load(self, datasource = datasource, *args, **kwargs):
			self.callback_loaded()
			return True
		return False
	
	def is_saved(self):
		
		return self._saved
	
	def __del__(self):
		
		if self._keep_temp:
			return
		try:
			clear_temp_dir()
		except:
			pass




from deposit.datasource.abstract_datasource import AbstractDatasource

from deposit.utils.fnc_files import (get_temp_path)
import uuid

class Memory(AbstractDatasource):
	
	def __init__(self):
		
		AbstractDatasource.__init__(self)
		
		self._uuid = uuid.uuid4().hex
	
	def get_name(self):
		
		return "memory db"
	
	def is_valid(self):
		
		return True
	
	def can_create(self):
		
		return True
	
	def get_folder(self):
		
		return get_temp_path(self._uuid)
	
	def create(self, *args, **kwargs):
		
		return True
	
	def save(self, store, progress = None, *args, **kwargs):
		
		return True
	
	def load(self, store, progress = None, *args, **kwargs):
		
		store.clear()
		store.set_datasource(self)
		
		return True



from deposit.utils.fnc_files import (as_url, url_to_path)

import os

class AbstractExternalsource(object):
	
	def __init__(self):
		
		self._path = None
	
	def get_name(self):
		
		if self._path is not None:
			name = os.path.basename(self._path)
			if name:
				return name
		return self.__class__.__name__
	
	def get_url(self):
		
		if self._path is not None:
			return as_url(self._path)
		return None
	
	def get_path(self):
		
		return self._path
	
	def set_path(self, path = None, url = None):
		
		if (path is None) and (url is not None):
			path = url_to_path(url)
		if path is None:
			self._path = None
			return
		self._path = os.path.normpath(os.path.abspath(path))
	
	def get_header_data(self, get_header, n_cols):
		# get_header(col, user_role = False) = column_name or (class_name, descriptor_name) if user_role = True
		
		return [get_header(col) for col in range(n_cols)]
	
	def get_row_data(self, row, get_item, n_cols):
		# get_item(row, col) = deposit_gui.QueryItem
		
		data = []
		for col in range(n_cols):
			item = get_item(row, col)
			if item.is_geometry():
				data.append(item.value.wkt)
			elif item.is_resource():
				data.append(item.value.url)
			else:
				data.append(item.get_display_data())
		return data
	
	def sheets(self):
		# re-implement
		
		return []
	
	def column_count(self, sheet):
		# re-implement
		
		return 0
	
	def row_count(self, sheet):
		# re-implement
		
		return 0
	
	def column_name(self, sheet, column_idx):
		# re-implement
		
		pass
	
	def data(self, sheet, row_idx, column_idx):
		# re-implement
		
		pass
	
	def save(self, get_header, get_item, n_rows, n_cols, path = None, url = None, *args, **kwargs):
		# get_header(col, user_role = False) = column_name or (class_name, descriptor_name) if user_role = True
		# get_item(row, col) = deposit_gui.QueryItem
		#
		# re-implement
		
		pass
	
	def load(self, path = None, url = None, *args, **kwargs):
		# re-implement
		
		pass
	
	def __str__(self):
		
		return self.__class__.__name__




from deposit import __version__
from deposit.datasource.abstract_filesource import AbstractFileSource

import pickle
import copy
import sys
import os

class Pickle(AbstractFileSource):
	
	EXTENSION = "pickle"
	
	def save_data(self, store, resources, path):
		
		for obj in store.G.iter_objects_data():
			obj._store = None
		for cls in store.G.iter_classes_data():
			cls._store = None
		
		self.update_progress(1, 10, text = "Saving data")
		
		with open(path, "wb") as f:
			pickle.dump(dict(
				object_relation_graph = store.G.objects_to_pickle(), 
				class_relation_graph = store.G.classes_to_pickle(), 
				class_membership_graph = store.G.members_to_pickle(), 
				resources = resources,
				local_folder = store._local_folder,
				max_order = store._max_order,
				user_tools = store._user_tools,
				queries = store._queries,
				deposit_version = __version__,
			), f, pickle.HIGHEST_PROTOCOL)
		
		for obj in store.G.iter_objects_data():
			obj._store = store
		for cls in store.G.iter_classes_data():
			cls._store = store
		
		self.update_progress(10)
		
		return True
	
	def load_data(self, path):
		
		self.update_progress(1, 10)
		with open(path, "rb") as f:
			data = pickle.load(f)
		return data
	
	def data_to_store(self, data, store):
		
		store.clear()
		store.G.objects_from_pickle(data["object_relation_graph"])
		store.G.classes_from_pickle(data["class_relation_graph"])
		store.G.members_from_pickle(data["class_membership_graph"])
		store._resources = data["resources"]
		store._local_folder = data["local_folder"]
		store._max_order = data["max_order"]
		store._user_tools = data["user_tools"]
		store._queries = data["queries"]
		
		for obj in store.G.iter_objects_data():
			obj._store = store
		for cls in store.G.iter_classes_data():
			cls._store = store
		
		self.update_progress(10)
		
		return True

